@@ -10,12 +10,12 @@ from filters import *
 from lxml import objectify
 from lxml import etree
 import base64
-import Image
+from PIL import Image
 from StringIO import StringIO
 import math
 import re
 import copy
- 
+
 SVG_VERSION = "1.1"
 SVG_NS      = "http://www.w3.org/2000/svg"
 XLINK_NS    = "http://www.w3.org/1999/xlink"
@@ -35,38 +35,38 @@ JOIN_STYLE = {
     1 : 'bevel',
     2 : 'miter'
 }
-                
+
 class DefaultShapeExporter(object):
     """
-    The default (abstract) Shape exporter class. 
+    The default (abstract) Shape exporter class.
     All shape exporters should extend this class.
-    
-    
+
+
     """
     def __init__(self, swf=None, debug=False, force_stroke=False):
         self.swf = None
         self.debug = debug
         self.force_stroke = force_stroke
-        
+
     def begin_bitmap_fill(self, bitmap_id, matrix=None, repeat=False, smooth=False):
         pass
     def begin_fill(self, color, alpha=1.0):
         pass
-    def begin_gradient_fill(self, type, colors, alphas, ratios, 
-                            matrix=None, 
-                            spreadMethod=SpreadMethod.PAD, 
+    def begin_gradient_fill(self, type, colors, alphas, ratios,
+                            matrix=None,
+                            spreadMethod=SpreadMethod.PAD,
                             interpolationMethod=InterpolationMethod.RGB,
                             focalPointRatio=0.0):
         pass
     def line_style(self,
-                    thickness=float('nan'), color=0, alpha=1.0, 
-                    pixelHinting=False, 
-                    scaleMode=LineScaleMode.NORMAL, 
-                    startCaps=None, endCaps=None, 
+                    thickness=float('nan'), color=0, alpha=1.0,
+                    pixelHinting=False,
+                    scaleMode=LineScaleMode.NORMAL,
+                    startCaps=None, endCaps=None,
                     joints=None, miterLimit=3.0):
         pass
     def line_gradient_style(self,
-                    thickness=float('nan'), color=0, alpha=1.0, 
+                    thickness=float('nan'), color=0, alpha=1.0,
                     pixelHinting=False,
                     scaleMode=LineScaleMode.NORMAL,
                     startCaps=None, endCaps=None,
@@ -79,15 +79,15 @@ class DefaultShapeExporter(object):
         pass
     def line_bitmap_style(self,
                     thickness=float('nan'),
-                    pixelHinting=False, 
-                    scaleMode=LineScaleMode.NORMAL, 
-                    startCaps=None, endCaps=None, 
+                    pixelHinting=False,
+                    scaleMode=LineScaleMode.NORMAL,
+                    startCaps=None, endCaps=None,
                     joints=None, miterLimit = 3.0,
                     bitmap_id=None, matrix=None, repeat=False, smooth=False):
-        pass   
+        pass
     def end_fill(self):
         pass
-            
+
     def begin_fills(self):
         pass
     def end_fills(self):
@@ -96,12 +96,12 @@ class DefaultShapeExporter(object):
         pass
     def end_lines(self):
         pass
-            
+
     def begin_shape(self):
         pass
     def end_shape(self):
         pass
-        
+
     def move_to(self, x, y):
         #print "move_to", x, y
         pass
@@ -120,13 +120,13 @@ class DefaultSVGShapeExporter(DefaultShapeExporter):
         self._e = objectify.ElementMaker(annotate=False,
                         namespace=SVG_NS, nsmap={None : SVG_NS, "xlink" : XLINK_NS})
         super(DefaultSVGShapeExporter, self).__init__()
-    
+
     def move_to(self, x, y):
         self.current_draw_command = ""
         self.path_data += "M" + \
             str(NumberUtils.round_pixels_20(x)) + " " + \
             str(NumberUtils.round_pixels_20(y)) + " "
-        
+
     def line_to(self, x, y):
         if self.current_draw_command != "L":
             self.current_draw_command = "L"
@@ -134,7 +134,7 @@ class DefaultSVGShapeExporter(DefaultShapeExporter):
         self.path_data += "" + \
             str(NumberUtils.round_pixels_20(x)) + " " + \
             str(NumberUtils.round_pixels_20(y)) + " "
-        
+
     def curve_to(self, cx, cy, ax, ay):
         if self.current_draw_command != "Q":
             self.current_draw_command = "Q"
@@ -144,47 +144,47 @@ class DefaultSVGShapeExporter(DefaultShapeExporter):
             str(NumberUtils.round_pixels_20(cy)) + " " + \
             str(NumberUtils.round_pixels_20(ax)) + " " + \
             str(NumberUtils.round_pixels_20(ay)) + " "
-        
+
     def begin_bitmap_fill(self, bitmap_id, matrix=None, repeat=False, smooth=False):
         self.finalize_path()
-        
+
     def begin_fill(self, color, alpha=1.0):
         self.finalize_path()
-    
+
     def end_fill(self):
         pass
         #self.finalize_path()
-    
+
     def begin_fills(self):
         pass
     def end_fills(self):
         self.finalize_path()
-            
-    def begin_gradient_fill(self, type, colors, alphas, ratios, 
-                            matrix=None, 
-                            spreadMethod=SpreadMethod.PAD, 
+
+    def begin_gradient_fill(self, type, colors, alphas, ratios,
+                            matrix=None,
+                            spreadMethod=SpreadMethod.PAD,
                             interpolationMethod=InterpolationMethod.RGB,
                             focalPointRatio=0.0):
         self.finalize_path()
-        
+
     def line_style(self,
-                    thickness=float('nan'), color=0, alpha=1.0, 
-                    pixelHinting=False, 
-                    scaleMode=LineScaleMode.NORMAL, 
-                    startCaps=None, endCaps=None, 
+                    thickness=float('nan'), color=0, alpha=1.0,
+                    pixelHinting=False,
+                    scaleMode=LineScaleMode.NORMAL,
+                    startCaps=None, endCaps=None,
                     joints=None, miterLimit=3.0):
         self.finalize_path()
-                       
+
     def end_lines(self):
         self.finalize_path()
-    
+
     def end_shape(self):
         self.finalize_path()
-            
+
     def finalize_path(self):
         self.current_draw_command = ""
         self.path_data = ""
-        
+
 class SVGShapeExporter(DefaultSVGShapeExporter):
     def __init__(self):
         self.path = None
@@ -195,10 +195,10 @@ class SVGShapeExporter(DefaultSVGShapeExporter):
         self.paths = {}
         self.fills_ended = False
         super(SVGShapeExporter, self).__init__()
-        
+
     def begin_shape(self):
         self.g = self._e.g()
-         
+
     def begin_fill(self, color, alpha=1.0):
         self.finalize_path()
         self.path.set("fill", ColorUtils.to_rgb_string(color))
@@ -209,19 +209,19 @@ class SVGShapeExporter(DefaultSVGShapeExporter):
             self.path.set("stroke-width", "1")
         else:
             self.path.set("stroke", "none")
-            
-    def begin_gradient_fill(self, type, colors, alphas, ratios, 
-                            matrix=None, 
-                            spreadMethod=SpreadMethod.PAD, 
+
+    def begin_gradient_fill(self, type, colors, alphas, ratios,
+                            matrix=None,
+                            spreadMethod=SpreadMethod.PAD,
                             interpolationMethod=InterpolationMethod.RGB,
                             focalPointRatio=0.0):
         self.finalize_path()
         gradient_id = self.export_gradient(type, colors, alphas, ratios, matrix, spreadMethod, interpolationMethod, focalPointRatio)
-        self.path.set("stroke", "none")    
+        self.path.set("stroke", "none")
         self.path.set("fill", "url(#%s)" % gradient_id)
-        
-    def export_gradient(self, type, colors, alphas, ratios, 
-                        matrix=None, 
+
+    def export_gradient(self, type, colors, alphas, ratios,
+                        matrix=None,
                         spreadMethod=SpreadMethod.PAD,
                         interpolationMethod=InterpolationMethod.RGB,
                         focalPointRatio=0.0):
@@ -230,7 +230,7 @@ class SVGShapeExporter(DefaultSVGShapeExporter):
         gradient = self._e.linearGradient() if type == GradientType.LINEAR \
             else self._e.radialGradient()
         gradient.set("gradientUnits", "userSpaceOnUse")
-        
+
         if type == GradientType.LINEAR:
             gradient.set("x1", "-819.2")
             gradient.set("x2", "819.2")
@@ -241,21 +241,21 @@ class SVGShapeExporter(DefaultSVGShapeExporter):
             if focalPointRatio < 0.0 or focalPointRatio > 0.0:
                 gradient.set("fx", str(819.2 * focalPointRatio))
                 gradient.set("fy", "0")
-        
+
         if spreadMethod == SpreadMethod.PAD:
             gradient.set("spreadMethod", "pad")
         elif spreadMethod == SpreadMethod.REFLECT:
             gradient.set("spreadMethod", "reflect")
         elif spreadMethod == SpreadMethod.REPEAT:
             gradient.set("spreadMethod", "repeat")
-        
+
         if interpolationMethod == InterpolationMethod.LINEAR_RGB:
             gradient.set("color-interpolation", "linearRGB")
-            
+
         if matrix is not None:
             sm = _swf_matrix_to_svg_matrix(matrix)
             gradient.set("gradientTransform", sm);
-        
+
         for i in range(0, len(colors)):
             entry = self._e.stop()
             offset = ratios[i] / 255.0
@@ -265,7 +265,7 @@ class SVGShapeExporter(DefaultSVGShapeExporter):
             if alphas[i] != 1.0:
                 entry.set("stop-opacity", str(alphas[i]))
             gradient.append(entry)
-        
+
         # prevent same gradient in <defs />
         key = etree.tostring(gradient)
         if key in self._gradients:
@@ -275,9 +275,9 @@ class SVGShapeExporter(DefaultSVGShapeExporter):
             self._gradient_ids[key] = gradient_id
             gradient.set("id", gradient_id)
             self.defs.append(gradient)
-            
+
         return gradient_id
-        
+
     def export_pattern(self, bitmap_id, matrix, repeat=False, smooth=False):
         self.num_patterns += 1
         bitmap_id = "c%d" % bitmap_id
@@ -299,20 +299,20 @@ class SVGShapeExporter(DefaultSVGShapeExporter):
         use.set(XLINK_HREF, "#%s" % bitmap_id)
         pattern.append(use)
         self.defs.append(pattern)
-        
+
         return pattern_id
-        
+
     def begin_bitmap_fill(self, bitmap_id, matrix=None, repeat=False, smooth=False):
         self.finalize_path()
         pattern_id = self.export_pattern(bitmap_id, matrix, repeat, smooth)
-        self.path.set("stroke", "none")    
+        self.path.set("stroke", "none")
         self.path.set("fill", "url(#%s)" % pattern_id)
-         
+
     def line_style(self,
-                    thickness=float('nan'), color=0, alpha=1.0, 
-                    pixelHinting=False, 
-                    scaleMode=LineScaleMode.NORMAL, 
-                    startCaps=None, endCaps=None, 
+                    thickness=float('nan'), color=0, alpha=1.0,
+                    pixelHinting=False,
+                    scaleMode=LineScaleMode.NORMAL,
+                    startCaps=None, endCaps=None,
                     joints=None, miterLimit=3.0):
         self.finalize_path()
         self.path.set("fill", "none")
@@ -325,14 +325,14 @@ class SVGShapeExporter(DefaultSVGShapeExporter):
 
     def line_gradient_style(self,
                     thickness=float('nan'),
-                    pixelHinting = False, 
-                    scaleMode=LineScaleMode.NORMAL, 
-                    startCaps=0, endCaps=0, 
+                    pixelHinting = False,
+                    scaleMode=LineScaleMode.NORMAL,
+                    startCaps=0, endCaps=0,
                     joints=0, miterLimit=3.0,
                     type = 1,
                     colors = [],
                     alphas = [],
-                    ratios = [], 
+                    ratios = [],
                     matrix=None,
                     spreadMethod=SpreadMethod.PAD,
                     interpolationMethod=InterpolationMethod.RGB,
@@ -346,12 +346,12 @@ class SVGShapeExporter(DefaultSVGShapeExporter):
         thickness = 1 if math.isnan(thickness) else thickness
         thickness = MINIMUM_STROKE_WIDTH if thickness < MINIMUM_STROKE_WIDTH else thickness
         self.path.set("stroke-width", str(thickness))
-    
+
     def line_bitmap_style(self,
                     thickness=float('nan'),
-                    pixelHinting=False, 
-                    scaleMode=LineScaleMode.NORMAL, 
-                    startCaps=None, endCaps=None, 
+                    pixelHinting=False,
+                    scaleMode=LineScaleMode.NORMAL,
+                    startCaps=None, endCaps=None,
                     joints=None, miterLimit = 3.0,
                     bitmap_id=None, matrix=None, repeat=False, smooth=False):
         self.finalize_path()
@@ -363,13 +363,13 @@ class SVGShapeExporter(DefaultSVGShapeExporter):
         thickness = 1 if math.isnan(thickness) else thickness
         thickness = MINIMUM_STROKE_WIDTH if thickness < MINIMUM_STROKE_WIDTH else thickness
         self.path.set("stroke-width", str(thickness))
-        
+
     def begin_fills(self):
         self.fills_ended = False
     def end_fills(self):
         self.finalize_path()
         self.fills_ended = True
-                                                         
+
     def finalize_path(self):
         if self.path is not None and len(self.path_data) > 0:
             self.path_data = self.path_data.rstrip()
@@ -388,17 +388,17 @@ class BaseExporter(object):
         self.force_stroke = force_stroke
         if swf is not None:
             self.export(swf)
-            
+
     def export(self, swf, force_stroke=False):
         self.force_stroke = force_stroke
         self.export_define_shapes(swf.tags)
         self.export_display_list(self.get_display_tags(swf.tags))
-        
+
     def export_define_bits(self, tag):
         png_buffer = StringIO()
         image = None
         if isinstance(tag, TagDefineBitsJPEG3):
-            
+
             tag.bitmapData.seek(0)
             tag.bitmapAlphaData.seek(0, 2)
             num_alpha = tag.bitmapAlphaData.tell()
@@ -430,21 +430,21 @@ class BaseExporter(object):
                 image = Image.open(buff)
             else:
                 image = Image.open(tag.bitmapData)
-            
+
         self.export_image(tag, image)
-    
+
     def export_define_bits_lossless(self, tag):
         image = Image.open(tag.bitmapData)
         self.export_image(tag, image)
-        
+
     def export_define_sprite(self, tag, parent=None):
         display_tags = self.get_display_tags(tag.tags)
         self.export_display_list(display_tags, parent)
-            
+
     def export_define_shape(self, tag):
         self.shape_exporter.debug = isinstance(tag, TagDefineShape4)
         tag.shapes.export(self.shape_exporter)
-        
+
     def export_define_shapes(self, tags):
         for tag in tags:
             if isinstance(tag, SWFTimelineContainer):
@@ -459,18 +459,18 @@ class BaseExporter(object):
                 self.export_define_bits(tag)
             elif isinstance(tag, TagDefineBitsLossless):
                 self.export_define_bits_lossless(tag)
-                  
+
     def export_display_list(self, tags, parent=None):
         self.clip_depth = 0
         for tag in tags:
             self.export_display_list_item(tag, parent)
-    
+
     def export_display_list_item(self, tag, parent=None):
         pass
-    
+
     def export_image(self, tag, image=None):
         pass
-        
+
     def get_display_tags(self, tags, z_sorted=True):
         dp_tuples = []
         for tag in tags:
@@ -484,20 +484,20 @@ class BaseExporter(object):
         for item in dp_tuples:
             display_tags.append(item[0])
         return display_tags
-    
+
     def serialize(self):
         return None
-        
+
 class SVGExporter(BaseExporter):
     def __init__(self, swf=None, margin=0):
         self._e = objectify.ElementMaker(annotate=False,
                         namespace=SVG_NS, nsmap={None : SVG_NS, "xlink" : XLINK_NS})
         self._margin = margin
         super(SVGExporter, self).__init__(swf)
-        
+
     def export(self, swf, force_stroke=False):
         """ Exports the specified SWF to SVG.
-        
+
         @param swf  The SWF.
         @param force_stroke Whether to force strokes on non-stroked fills.
         """
@@ -509,10 +509,10 @@ class SVGExporter(BaseExporter):
         self.svg.append(self.root)
         self.shape_exporter.defs = self.defs
         self._num_filters = 0
-        
+
         # GO!
         super(SVGExporter, self).export(swf, force_stroke)
-        
+
         # Setup svg @width, @height and @viewBox
         # and add the optional margin
         self.bounds = SVGBounds(self.svg)
@@ -520,36 +520,36 @@ class SVGExporter(BaseExporter):
         self.svg.set("height", "%dpx" % round(self.bounds.height))
         if self._margin > 0:
             self.bounds.grow(self._margin)
-        vb = [self.bounds.minx, self.bounds.miny, 
+        vb = [self.bounds.minx, self.bounds.miny,
               self.bounds.width, self.bounds.height]
         self.svg.set("viewBox", "%s" % " ".join(map(str,vb)))
-        
+
         # Return the SVG as StringIO
         return self._serialize()
-    
+
     def _serialize(self):
         return StringIO(etree.tostring(self.svg,
                 encoding="UTF-8", xml_declaration=True))
-                
+
     def export_define_sprite(self, tag, parent=None):
         id = "c%d"%tag.characterId
         g = self._e.g(id=id)
         self.defs.append(g)
         self.clip_depth = 0
         super(SVGExporter, self).export_define_sprite(tag, g)
-            
+
     def export_define_shape(self, tag):
         self.shape_exporter.force_stroke = self.force_stroke
         super(SVGExporter, self).export_define_shape(tag)
         shape = self.shape_exporter.g
         shape.set("id", "c%d" % tag.characterId)
         self.defs.append(shape)
-    
+
     def export_display_list_item(self, tag, parent=None):
         g = self._e.g()
         use = self._e.use()
         is_mask = False
-        
+
         if tag.hasMatrix:
             use.set("transform", _swf_matrix_to_svg_matrix(tag.matrix))
         if tag.hasClipDepth:
@@ -572,7 +572,7 @@ class SVGExporter(BaseExporter):
 
         if tag.hasColorTransform:
             filter_cxform = self.export_color_transform(tag.colorTransform, svg_filter)
-            filters.append(filter_cxform) 
+            filters.append(filter_cxform)
         if tag.hasFilterList and len(tag.filters) > 0:
             cxform = "color-xform" if tag.hasColorTransform else None
             f = self.export_filters(tag, svg_filter, cxform)
@@ -581,10 +581,10 @@ class SVGExporter(BaseExporter):
         if tag.hasColorTransform or (tag.hasFilterList and len(filters) > 0):
             self.defs.append(svg_filter)
             use.set("filter", "url(#%s)" % filter_id)
-    
+
         use.set(XLINK_HREF, "#c%s" % tag.characterId)
         g.append(use)
-        
+
         if is_mask:
             self.defs.append(g)
         else:
@@ -593,22 +593,22 @@ class SVGExporter(BaseExporter):
             else:
                 self.root.append(g)
         return use
-        
+
     def export_color_transform(self, cxform, svg_filter, result='color-xform'):
         fe_cxform = self._e.feColorMatrix()
         fe_cxform.set("in", "SourceGraphic")
         fe_cxform.set("type", "matrix")
         fe_cxform.set("values", " ".join(map(str, cxform.matrix)))
         fe_cxform.set("result", "cxform")
-        
+
         fe_composite = self._e.feComposite(operator="in")
         fe_composite.set("in2", "SourceGraphic")
         fe_composite.set("result", result)
-        
+
         svg_filter.append(fe_cxform)
         svg_filter.append(fe_composite)
         return result
-    
+
     def export_filters(self, tag, svg_filter, cxform=None):
         num_filters = len(tag.filters)
         elements = []
@@ -641,16 +641,16 @@ class SVGExporter(BaseExporter):
             else:
                 raise Exception("unknown filter: ", swf_filter)
         return elements
-    
+
 #   <filter id="test-filter" x="-50%" y="-50%" width="200%" height="200%">
 #		<feGaussianBlur in="SourceAlpha" stdDeviation="6" result="blur"/>
 #		<feOffset dy="0" dx="0"/>
 #		<feComposite in2="SourceAlpha" operator="arithmetic"
 #			k2="-1" k3="1" result="shadowDiff"/>
 #		<feFlood flood-color="black" flood-opacity="1"/>
-#		<feComposite in2="shadowDiff" operator="in"/>	
+#		<feComposite in2="shadowDiff" operator="in"/>
 #	</filter>;
-        			
+
     def export_filter_dropshadow(self, swf_filter, svg_filter, blend_in=None, result="offsetBlur"):
         gauss = self._e.feGaussianBlur()
         gauss.set("in", "SourceAlpha")
@@ -670,15 +670,15 @@ class SVGExporter(BaseExporter):
             svg_filter.append(flood)
             svg_filter.append(composite1)
         else:
-            SVGFilterFactory.create_drop_shadow_filter(self._e, svg_filter, 
-                None, 
-                swf_filter.blurX/20.0, 
+            SVGFilterFactory.create_drop_shadow_filter(self._e, svg_filter,
+                None,
+                swf_filter.blurX/20.0,
                 swf_filter.blurY/20.0,
                 blend_in,
                 result)
         #print etree.tostring(svg_filter, pretty_print=True)
         return result
-        
+
     def export_image(self, tag, image=None):
         if image is not None:
             buff = StringIO()
@@ -707,7 +707,7 @@ class SVGFilterFactory(object):
         filter.append(offset)
         filter.append(blend)
         return result
-        
+
     @classmethod
     def export_color_matrix_filter(cls, e, filter, matrix, svg_filter, attr_in=None, result='color-matrix'):
         attr_in = "SourceGraphic" if attr_in is None else attr_in
@@ -719,14 +719,14 @@ class SVGFilterFactory(object):
         filter.append(fe_cxform)
         #print etree.tostring(filter, pretty_print=True)
         return result
-    
+
     @classmethod
     def export_glow_filter(cls, e, filter, attr_in=None, result="glow-out"):
         attr_in = "SourceGraphic" if attr_in is None else attr_in
         gaussianBlur = SVGFilterFactory.create_gaussian_blur(e, attr_in=attr_in, attr_deviaton="1", result=result)
         filter.append(gaussianBlur)
         return result
-        
+
     @classmethod
     def create_blend(cls, e, attr_in=None, attr_in2="BackgroundImage", mode="normal", result=None):
         blend = e.feBlend()
@@ -737,7 +737,7 @@ class SVGFilterFactory(object):
         if result is not None:
             blend.set("result", result)
         return blend
-        
+
     @classmethod
     def create_gaussian_blur(cls, e, attr_in="SourceAlpha", attr_deviaton="3", result=None):
         gaussianBlur = e.feGaussianBlur()
@@ -746,7 +746,7 @@ class SVGFilterFactory(object):
         if result is not None:
             gaussianBlur.set("result", result)
         return gaussianBlur
-        
+
     @classmethod
     def create_offset(cls, e, attr_in=None, dx=0, dy=0, result=None):
         offset = e.feOffset()
@@ -757,7 +757,7 @@ class SVGFilterFactory(object):
         if result is not None:
             offset.set("result", result)
         return offset
-        
+
 class SVGBounds(object):
     def __init__(self, svg=None):
         self.minx = 1000000.0
@@ -874,12 +874,12 @@ class SVGBounds(object):
 
 def _encode_jpeg(data):
     return "data:image/jpeg;base64," + base64.encodestring(data)[:-1]
-                                        
+
 def _encode_png(data):
     return "data:image/png;base64," + base64.encodestring(data)[:-1]
 
 def _swf_matrix_to_matrix(swf_matrix=None, need_scale=False, need_translate=True, need_rotation=False, unit_div=20.0):
-    
+
     if swf_matrix is None:
         values = [1,0,0,0,0,1,0,0,0,0,1,0,0,0,0,1]
     else:
@@ -893,9 +893,9 @@ def _swf_matrix_to_matrix(swf_matrix=None, need_scale=False, need_translate=True
         if need_translate:
             values[4] /= unit_div
             values[5] /= unit_div
-        
+
     return values
-    
+
 def _swf_matrix_to_svg_matrix(swf_matrix=None, need_scale=False, need_translate=True, need_rotation=False, unit_div=20.0):
     values = _swf_matrix_to_matrix(swf_matrix, need_scale, need_translate, need_rotation, unit_div)
     str_values = ",".join(map(str, values))
